"""

"""

import cv2

from videopypeline import core


def flatten(iterable):
    for item in iterable:
        yield item


def read_video_file(filepath):
    cap = cv2.VideoCapture(filepath)
    assert cap.isOpened(), filepath

    while cap.grab():
        yield cap.retrieve()[1]  # BGR-format

    cap.release()


def constant_generator(c):
    while True:
        yield c


class Flatten(core.Generator):
    def __init__(self, iterable, **kwargs):
        super().__init__(lambda: flatten(iterable), **kwargs)


class ReadVideoFile(core.Generator):
    def __init__(self, filepath, **kwargs):
        super().__init__(lambda: read_video_file(filepath), **kwargs)


class Constant(core.Generator):
    def __init__(self, c, **kwargs):
        super().__init__(lambda: constant_generator(c), **kwargs)



r"""
This module contains the core functions and classes for the videopypeline package.
"""

from __future__ import annotations

import random
import typing
import time

import graphviz
import numpy as np


class AbortPipeline(Exception):
    """ Exception used to interrupt the pipeline execution.

    This Exception is used by :py:class:`core.Function`-Nodes. When the
    :py:attr:`core.Function.filter_fn`-Method returns False this exception is thrown.
    """
    pass


class AbstractNode:
    """Base node class.

    :param process_fn: The operation which will be wrapped by the node
    :type process_fn: callable
    :param name: The name of the node
    :type name: str, optional
    :param verbose: If true, prints information, defaults to False
    :type verbose: bool, optional
    :param debug_verbose: If true, prints debug information, defaults to False
    :type debug_verbose: bool, optional
    :param aggregate: If true, invokes the previous nodes until their generator exhausts. The output of every invocation
        is stored in a list before passing it to subsequent nodes (See ``collect``). Defaults to False.
    :type aggregate: bool, optional
    :param collect: If true, when aggregating, the nodes output will be stored, otherwise discarded (See ``aggregate``).
        Defaults to False.
    :type collect: bool, optional
    :param timeit: If true, the execution time for invoking this node is timed, defaults to False
    :type timeit: bool, optional
    """

    def __init__(self, process_fn: typing.Callable, name: str = "", verbose: bool = False, debug_verbose: bool = False,
                 aggregate: bool = False, collect: bool = True, timeit: bool = False):
        assert callable(process_fn)
        self.previous: typing.List[AbstractNode] = []  #: The previous linked nodes
        self.process_fn: typing.Callable = process_fn  #: The wrapped callable
        self.cache_data = None  #: The cached output of this node
        self.time_data = []  #: The collected timer data

        self.name: str = name  #: The name of the node
        self.verbose: bool = bool(verbose)  #: If true, console output is more verbose
        self.debug_verbose: bool = bool(debug_verbose)  #: If true, debug information will be printed
        self.aggregate: bool = bool(aggregate)  #: If true, :py:meth:`__call__` invokes :py:meth:`invoke` until the generator exhausts, otherwise only once
        self.collect: bool = bool(collect)  #: Applicable if :py:attr:`aggregate` is true. If true, the outputs will be collected into a list, otherwise discared
        self.timeit: bool = bool(timeit)  #: If true, the execution time is timed.

    def __call__(self, *args):
        """ Wrapper for :py:meth:`infer` and :py:meth:`model`

        This method is a convenience wrapper for two functions :py:meth:`infer` and :py:meth:`model`.
        If ``args[0]`` is of type ``AbstractNode`` or ``List[AbstractNode]``, :py:meth:`model` is invoked, otherwise
        :py:meth:`infer`.

        There are two types of inferring. This is controlled by :py:attr:`aggregate`. The two types are defined as
        follows:

        1. (``aggregate == False``) :py:meth:`infer` is called once and the output is returned.
        2. (``aggregate == True``) :py:meth:`infer` is called until the underlying generator exhausts.
           If :py:attr:`collect` is ``True`` the output for every iteration is stored in a list which is returned.
           If :py:attr:`collect` is ``True`` the output is discarded.

        :param args: asdf
        :return: asdf
        """
        is_modelling = self.is_modelling(*args)

        if is_modelling:  # model
            return self.model(args[0])

        elif not is_modelling and not self.aggregate:  # infer once
            return self.infer()

        elif not is_modelling and self.aggregate:  # infer until generator exhausts
            collection, iteration, run = [], 0, True

            try:
                while run:
                    try:
                        if self.verbose:
                            print(f"Aggregating {iteration}")

                        output = self.infer()
                        self.clear_cache()

                        if self.collect:
                            collection.append(output)
                    except AbortPipeline:
                        self.clear_cache()
                    except StopIteration:
                        run = False

                    iteration += 1
            except KeyboardInterrupt:
                if self.verbose:
                    print("Interrupted...")

            return collection
            
        else:
            assert False

    def __getitem__(self, n: int) -> AbstractNode:
        """

        :param n:
        :return:
        """
        assert isinstance(n, int)
        return AbstractNode(lambda *args: args[0][n], name=f"ArgSelect{n}")(self)

    def infer(self):
        """ Infer

        :return: asdf
        """
        assert isinstance(self.previous, list)
        assert all(isinstance(p, AbstractNode) for p in self.previous)

        # Return cached result if available
        if self.cache_data is not None:
            return self.cache_data

        # Infer previous nodes
        previous_output = [prev.infer() for prev in self.previous]

        if self.debug_verbose:
            size = previous_output.shape if isinstance(previous_output, np.ndarray) else ''
            print(f'Input: {type(previous_output)}{size}')

        t0 = None
        if self.timeit:
            t0 = time.perf_counter()

        # Infer current node
        self.cache_data = self.process_fn(*previous_output)

        if self.timeit:
            t1 = time.perf_counter()
            self.time_data.append(t1 - t0)

        if self.debug_verbose:
            size = self.cache_data.shape if isinstance(self.cache_data, np.ndarray) else ''
            print(f'Output: {type(self.cache_data)}{size}')

        return self.cache_data

    def clear_cache(self):
        """

        :return:
        """
        for p in self.previous:
            p.clear_cache()
        self.cache_data = None

    def model(self, node: AbstractNode) -> AbstractNode:
        """

        :param node:
        :return:
        """
        assert self.is_modelling(node)

        if isinstance(node, list):
            assert all(isinstance(n, AbstractNode) for n in node)
            self.previous.extend(node)
        else:
            assert isinstance(node, AbstractNode)
            self.previous.append(node)
        return self

    def start(self):
        """

        :return:
        """
        for p in self.previous:
            p.start()
        self.start_callback()

    def end(self):
        """

        :return:
        """
        for p in self.previous:
            p.end()
        self.end_callback()

    def start_callback(self):
        """

        :return:
        """
        pass

    def end_callback(self):
        """

        :return:
        """
        pass

    @staticmethod
    def is_modelling(*args) -> bool:
        """

        :param args:
        :return:
        """
        if len(args) != 1:
            return False

        node = args[0]
        one_parent = isinstance(node, AbstractNode)
        many_parents = isinstance(node, list) and all(isinstance(n, AbstractNode) for n in node)
        return one_parent or many_parents


class Function(AbstractNode):
    """

    """

    def __init__(self, process_fn, **kwargs):
        super().__init__(process_fn, **kwargs)


class Generator(Function):
    """

    """

    def __init__(self, generator_fn: typing.Callable[[], typing.Generator], **kwargs):
        super().__init__(self.generate, **kwargs)
        self.generator = generator_fn()

    def generate(self):
        return next(self.generator)


class Action(Function):
    """

    """

    def __init__(self, action_fn: typing.Callable, **kwargs):
        super().__init__(self.action, **kwargs)
        self.action_fn = action_fn

    def action(self, *args):
        self.action_fn(*args)
        return args[0] if len(args) == 1 else args


class Filter(Action):
    """

    """

    def __init__(self, filter_fn: typing.Callable[..., bool], **kwargs):
        super().__init__(self.filter, **kwargs)
        self.filter_fn = filter_fn

    def filter(self, *args):
        if not self.filter_fn(*args):
            raise AbortPipeline()


class Pipeline(Function):
    """

    """

    def __init__(self, end_node: AbstractNode | typing.List[AbstractNode], **kwargs):
        super().__init__(self.pipeline, **kwargs)
        if isinstance(end_node, list):
            assert all(isinstance(n, AbstractNode) for n in end_node)
            assert len(end_node[0].previous) == 0

            for i in range(len(end_node) - 1):
                end_node[i + 1](end_node[i])

            self.end_node: AbstractNode = end_node[-1]
        elif isinstance(end_node, AbstractNode):
            self.end_node: AbstractNode = end_node
        else:
            assert False, type(end_node)

        self.previous = [end_node]

    def pipeline(self, *args):
        self.end_node.start()
        ret = self.end_node(*args)
        self.end_node.end()
        return ret

    def render_model(self):
        def get_name(n):
            def rnd_hex():
                hex_chars = [c for c in "ABCDEF0123456789"]
                return ''.join([random.choice(hex_chars) for _ in range(5)])

            return (n.__class__.__name__ if n.name == "" else n.name) + "-" + rnd_hex()

        dot = graphviz.Digraph('pipeline' if self.name == "" else self.name, format='png')

        graph = self.traverse_dfs(self.end_node)
        tr = {n: get_name(n) for n in graph.keys()}

        for node, previous in self.traverse_dfs(self.end_node).items():
            name = tr[node]
            dot.node(name)

            for prev in previous:
                prev_name = tr[prev]
                dot.edge(prev_name, name)

        # dot.render(outfile="filename.png", cleanup=True)
        return dot

    @staticmethod
    def traverse_dfs(node: AbstractNode) -> typing.Dict[AbstractNode, typing.List[AbstractNode]]:
        nodes = {}
        stack = [node]

        while stack:
            current = stack.pop()

            if current not in nodes:
                nodes[current] = []

                # if previous is empty then current must be a generator
                assert current.previous or isinstance(current, Generator) or isinstance(current, Pipeline)

                for prev in current.previous:
                    stack.append(prev)
                    nodes[current].append(prev)

        return nodes



import cv2
import numpy as np

from videopypeline import core


class VideoWriter(core.Action):

    def __init__(self, output_path, fps, **kwargs):
        super().__init__(self.write, **kwargs)
        self.output_path = output_path
        self.fps = fps
        self.writer = None

    def end_callback(self):
        if self.writer is not None:
            self.writer.release()

    def write(self, frame):
        frame = frame[0] if isinstance(frame, tuple) else frame
        assert isinstance(frame, np.ndarray), type(frame)
        if frame.ndim == 2:
            frame = np.dstack([frame, frame, frame])
        elif frame.ndim == 3:
            assert frame.shape[2] == 3
        else:
            assert False

        if self.writer is None:
            shape = tuple(reversed(frame.shape[:2]))
            fourcc = cv2.VideoWriter_fourcc('m', 'p', '4', 'v')
            self.writer = cv2.VideoWriter(self.output_path, fourcc, self.fps, shape)

        self.writer.write(frame.astype(np.uint8))


class VideoWriter2(core.Action):

    def __init__(self, output_path, fps):
        super().__init__(self.write)
        self.output_path = output_path
        self.fps = fps

    def write(self, frames):
        frames = frames[0] if isinstance(frames, tuple) else frames
        assert isinstance(frames, list)
        assert all(type(frame) == np.ndarray for frame in frames)
        if len(frames) == 0:
            return

        shape = list(reversed(frames[0].shape[:2]))
        writer = cv2.VideoWriter(self.output_path, cv2.VideoWriter_fourcc('m', 'p', '4', 'v'), self.fps, shape)

        for frame in frames:
            writer.write(frame.astype(np.uint8))

        writer.release()



"""

"""

import typing

import cv2
import numpy as np

from videopypeline import core


def _check_tuple(obj, true_len=2, true_type=int | np.int32):
    assert isinstance(obj, tuple), type(obj)
    assert len(obj) == true_len, len(obj)
    assert all(isinstance(o, true_type) for o in obj), [type(o) for o in obj]
    return True


def _check_nparray(array, ndim=None, dtype=None):
    assert isinstance(array, np.ndarray), type(array)
    assert ndim is None or array.ndim == ndim, array.ndim
    assert dtype is None or array.dtype == dtype, array.dtype
    return True


def crop(frame, position, size):
    assert _check_nparray(frame)
    assert _check_tuple(position)
    assert _check_tuple(size)
    return frame[position[0]:position[0]+size[0], position[1]:position[1]+size[1]]


def affine_transformation(frame, pos_d=None, rot_d=None, scale_d=None):
    assert True if pos_d is None else _check_tuple(pos_d, true_type=float | np.float)
    assert True if rot_d is None else isinstance(rot_d, float)
    assert True if scale_d is None else isinstance(scale_d, float)

    scale_d = 1 if scale_d is None else scale_d
    img_shape = frame.shape[1::-1]
    pivot_pos = np.array(img_shape) / 2 + pos_d

    rot_mat = cv2.getRotationMatrix2D(pivot_pos, rot_d, scale_d)
    result = cv2.warpAffine(frame, rot_mat, img_shape, flags=cv2.INTER_LINEAR)

    return result


def perspective_transformation(frame, src_points, dst_points, out_shape):
    assert _check_nparray(frame)
    assert isinstance(src_points, np.ndarray)
    assert isinstance(dst_points, np.ndarray)
    assert src_points.ndim == 2
    assert dst_points.ndim == 2
    assert _check_tuple(out_shape)

    mat = cv2.getPerspectiveTransform(src_points, dst_points)
    out_frame = cv2.warpPerspective(frame, mat, out_shape)
    return out_frame


def smooth(frame, window_size):
    assert _check_nparray(frame)
    assert isinstance(window_size, int)

    out_frame = np.zeros_like(frame)
    cv2.GaussianBlur(frame, (window_size, window_size), 0, out_frame, 0, cv2.BORDER_CONSTANT)

    return out_frame


def rgb_to_greyscale(frame):
    assert _check_nparray(frame, ndim=3)
    return np.array(0.0721 * frame[:, :, 0] + 0.7154 * frame[:, :, 1] + 0.2125 * frame[:, :, 2])  # BRG


def greyscale_to_rgb(frame):
    assert _check_nparray(frame, ndim=2)
    return np.dstack([frame, frame, frame])


def filter_largest_contour(contours):
    assert isinstance(contours, tuple)

    if len(contours) == 0:
        return None
    else:
        return max(contours, key=cv2.contourArea)


def get_contour_center(contour):
    if contour is None:
        return None
    else:
        assert _check_nparray(contour)
        mom = cv2.moments(contour)
        x, y = np.int32(mom["m10"] / mom["m00"]), np.int32(mom["m01"] / mom["m00"])
        return np.array([x, y])


def find_contours(frame):
    assert _check_nparray(frame, ndim=2)
    contours, _ = cv2.findContours(frame, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    return contours


def convex_contour(contour):
    if contour is None:
        return None
    else:
        assert _check_nparray(contour)
        hull = cv2.convexHull(contour)
        return hull


def bounding_box(contour):
    if contour is None:
        return None
    else:
        return cv2.boundingRect(contour)


def draw_contour_centers(frame, center):
    assert _check_nparray(frame)

    if center is None:
        return frame
    else:
        assert _check_nparray(center)

        out_frame = np.array(frame)
        cv2.circle(out_frame, tuple(center), 10, (255, 0, 255), -1)

        return out_frame


def draw_text(frame, text, org=(100, 100), scale=3, color=(255, 0, 255), thickness=3):
    assert _check_nparray(frame)
    font = cv2.FONT_HERSHEY_SIMPLEX
    return cv2.putText(np.array(frame), f'{text}', org, font, scale, color, thickness, cv2.LINE_AA)


def draw_line(frame, start_pos, end_pos, color, thickness=3):
    assert _check_nparray(frame, ndim=3)
    assert _check_tuple(start_pos)
    assert _check_tuple(end_pos)
    assert _check_tuple(color, true_len=3)
    assert isinstance(thickness, int)

    out_frame = np.array(frame)
    cv2.line(out_frame, (start_pos[0], start_pos[1]), (end_pos[0], end_pos[1]), color, thickness)

    return out_frame


def threshold(frame, t):
    assert _check_nparray(frame)
    assert isinstance(t, int)

    out_frame = np.zeros_like(frame)
    cv2.threshold(frame, t, frame.max(initial=0), cv2.THRESH_BINARY, out_frame)

    return out_frame


def erode(frame, kernel):
    assert _check_nparray(frame)
    assert _check_nparray(kernel, ndim=2)

    out_frame = np.zeros_like(frame)
    cv2.erode(frame, kernel, out_frame)

    return out_frame


def dilate(frame, kernel):
    assert _check_nparray(frame)
    assert _check_nparray(kernel, ndim=2)

    out_frame = np.zeros_like(frame)
    cv2.dilate(frame, kernel, out_frame)

    return out_frame


def canny_edge(frame, t1, t2):
    assert _check_nparray(frame, ndim=2)
    assert isinstance(t1, int)
    assert isinstance(t2, int)

    in_frame = frame.astype(np.uint8)
    out_frame = np.zeros_like(in_frame)
    cv2.Canny(in_frame, t1, t2, out_frame)

    return out_frame


def stack(rows, cols, *images):
    assert isinstance(rows, int)
    assert isinstance(cols, int)
    assert all(_check_nparray(i) for i in images)
    assert len(images) <= rows * cols, len(images)

    ref = images[0].shape
    out_image = np.zeros((ref[0] * rows, ref[1] * cols, 3))

    for i in range(rows):
        for j in range(cols):
            idx = i * cols + j
            if idx < len(images):
                img = images[idx]
                sub_img = img if img.ndim == 3 else np.dstack([img, img, img])
                out_image[i * ref[0]: (i+1) * ref[0], j * ref[1]: (j+1) * ref[1]] = sub_img

    return out_image


class Crop(core.Function):
    def __init__(self, position: typing.Tuple[int, int], size: typing.Tuple[int, int], **kwargs):
        super().__init__(lambda frame: crop(frame, position, size), **kwargs)


class AffineTransformation(core.Function):
    def __init__(self, pos_d=None, rot_d=None, scale_d=None, **kwargs):
        super().__init__(lambda frame: affine_transformation(frame, pos_d, rot_d, scale_d), **kwargs)


class PerspectiveTransformation(core.Function):
    def __init__(self, src_points, dst_points, out_shape, **kwargs):
        super().__init__(lambda frame: perspective_transformation(frame, src_points, dst_points, out_shape), **kwargs)


class Smooth(core.Function):
    def __init__(self, window_size: int, **kwargs):
        super().__init__(lambda frame: smooth(frame, window_size), **kwargs)


class Rgb2Greyscale(core.Function):
    def __init__(self, **kwargs):
        super().__init__(rgb_to_greyscale, **kwargs)


class Greyscale2Rgb(core.Function):
    def __init__(self, **kwargs):
        super().__init__(greyscale_to_rgb, **kwargs)


class FilterLargestContour(core.Function):
    def __init__(self, **kwargs):
        super().__init__(filter_largest_contour, **kwargs)


class GetContourCenter(core.Function):
    def __init__(self, **kwargs):
        super().__init__(get_contour_center, **kwargs)


class FindContours(core.Function):
    def __init__(self, **kwargs):
        super().__init__(lambda frame: find_contours(frame), **kwargs)


class ConvexHull(core.Function):
    def __init__(self, **kwargs):
        super().__init__(convex_contour, **kwargs)


class BoundingBox(core.Function):
    def __init__(self, **kwargs):
        super().__init__(bounding_box, **kwargs)


class DrawContourCenters(core.Function):
    def __init__(self, **kwargs):
        super().__init__(draw_contour_centers, **kwargs)


class DrawText(core.Function):
    def __init__(self, org=(100, 100), scale=3, color=(255, 0, 255), thickness=3, **kwargs):
        super().__init__(lambda frame, text: draw_text(frame, f'{text}', org=org, scale=scale, color=color, thickness=thickness), **kwargs)


class DrawMovementPath(core.Function):
    def __init__(self, color_coeff: int = 4, **kwargs):
        super().__init__(self.draw_movement_path, **kwargs)
        self.last_center = None
        self.lines = []
        self.color_coeff = color_coeff

    def draw_movement_path(self, frame, center):
        assert _check_nparray(frame)

        if frame.ndim == 2:
            frame = greyscale_to_rgb(frame)

        if center is not None and self.last_center is None:
            self.lines = []
        elif center is not None and self.last_center is not None:
            assert _check_nparray(center)
            self.lines.append((self.last_center, center))

        if len(self.lines) >= 2:
            lines = np.array(self.lines)
            last_centers, centers = lines[:, 0], lines[:, 1]

            for lc, c in zip(last_centers, centers):
                b = int(min(abs(c[0] - lc[0]) * self.color_coeff, 255))
                g = int(min(abs(c[1] - lc[1]) * self.color_coeff, 255))
                frame = draw_line(frame, tuple(lc), tuple(c), (b, g, 0))

        self.last_center = center
        return frame


class Threshold(core.Function):
    def __init__(self, t: int, **kwargs):
        super().__init__(lambda frame: threshold(frame, t), **kwargs)


class Erode(core.Function):
    def __init__(self, window_size: int, **kwargs):
        kernel = np.ones((window_size, window_size), 'uint8')
        super().__init__(lambda frame: erode(frame, kernel), **kwargs)


class Dilate(core.Function):
    def __init__(self, window_size: int, **kwargs):
        kernel = np.ones((window_size, window_size), 'uint8')
        super().__init__(lambda frame: dilate(frame, kernel), **kwargs)


class CannyEdge(core.Function):
    def __init__(self, t1: int, t2: int, **kwargs):
        super().__init__(lambda frame: canny_edge(frame, t1, t2), **kwargs)


class AbsDiff(core.Function):
    def __init__(self, **kwargs):
        super().__init__(self.abs_diff, **kwargs)
        self.last_frame = None

    def abs_diff(self, frame):
        assert _check_nparray(frame)

        diff = cv2.absdiff(frame, frame) if self.last_frame is None else cv2.absdiff(frame, self.last_frame)
        self.last_frame = frame

        return diff


class Stack(core.Function):
    def __init__(self, rows: int, cols: int, **kwargs):
        super().__init__(lambda *images: stack(rows, cols, *images), **kwargs)
